import { InjectionKey } from 'vue'
import { createStore, Store, createLogger } from 'vuex'
import { LOGIN_SUCCESS, LOGOUT, TITLE } from './mutation-types'
import { FETCH_ACCOUNT, FETCH_BUDGET, GET, REGISTER, IMPORT_YNAB, LOGIN, NEW_BUDGET, POST, SET_CURRENT_ACCOUNT, SET_CURRENT_BUDGET} from './action-types'

export interface State {
    Session: {
        Token?: string
        User?: string
    },
    ShowMenu?: boolean,
    Budgets: Map<string, Budget>,
    CurrentBudgetID?: string,
    Accounts: Map<string, Account>,
    CurrentAccountID?: string,
    Categories: [],
    Transactions: [],
    Assignments: []
}

export interface Budget {
    ID: string
}

export interface Account {
    ID: string
    OnBudget: boolean
}

export const key: InjectionKey<Store<State>> = Symbol()

export const store = createStore<State>({
    state: {
            Session: {
                Token: undefined,
                User: undefined
            },
            ShowMenu: undefined,
            Budgets: new Map<string, Budget>(),
            CurrentBudgetID: undefined,
            Accounts: new Map<string, Account>(),
            CurrentAccountID: undefined,
            Categories: [],
            Transactions: [],
            Assignments: []
    },
    mutations: {
        deleteBudget(state : State, budgetid : string) {
            state.Budgets.delete(budgetid)
        },
        toggleMenu(state) {
            state.ShowMenu = !state.ShowMenu;
        },
        initializeStore(state) {
            const store = localStorage.getItem("store");
            if (store) {
                this.replaceState(
                    Object.assign(state, JSON.parse(store))
                );
            }
        },
        [TITLE](state, title) {
            document.title = "Budgeteer - " + title;
        },
        [LOGIN_SUCCESS](state, result) {
            state.Session = {
                User: result.User,
                Token: result.Token
            };
            for (const budget of result.Budgets) {
                state.Budgets.set(budget.ID, budget)
            }
        },
        setBudgets(state, budgets) {
            state.Budgets = budgets;
        },
        addBudget(state, budget) {
            state.Budgets.set(budget.ID, budget);
        },
        [LOGOUT](state, token) {
            state.Session = { Token: undefined, User: undefined };
            state.Budgets.clear();
            state.Accounts.clear();
            state.Categories = [];
            state.Transactions = [];
            state.Assignments = [];
        },
        setCurrentBudgetID(state, budgetid) {
            state.CurrentBudgetID = budgetid;
        },
        setCurrentAccountID(state, accountid) {
            state.CurrentAccountID = accountid;
        },
        setAccounts(state, accounts) {
            state.Accounts = accounts;
        },
        setTransactions(state, transactions) {
            state.Transactions = transactions;
        }
    },
    actions: {
        [LOGIN]({ state, commit }, login) {
            return fetch("/api/v1/user/login", { method: "POST", body: JSON.stringify(login) })
                .then(x => x.json())
                .then(x => commit(LOGIN_SUCCESS, x))
        },
        [REGISTER]({ state, commit }, login) {
            return fetch("/api/v1/user/register", { method: "POST", body: JSON.stringify(login) })
                .then(x => x.json())
                .then(x => commit(LOGIN_SUCCESS, x))
        },
        [IMPORT_YNAB]({ getters, dispatch }, formData) {
            return dispatch("POST", { path: "/budget/" + getters.CurrentBudget.ID + "/import/ynab", body: formData });
        },
        [GET]({ getters }, { path }) {
            return fetch("/api/v1" + path, {
                headers: getters.AuthHeaders,
            })
        },
        [POST]({ getters }, { path, body }) {
            return fetch("/api/v1" + path, {
                method: "POST",
                headers: getters.AuthHeaders,
                body: body,
            })
        },
        /*async fetchDashboard ({state, commit, rootState}) {
            const response = await fetch("/api/v1/dashboard", {
                headers: {
                    'Authorization': 'Bearer ' + rootState.Session.Token
                }
            })
            const data = await response.json();
            commit("setBudgets", data.Budgets);
        },*/
        async [NEW_BUDGET]({ state, commit, dispatch, rootState }, budgetName) {
            const result = await dispatch("POST", {
                path: "/budget/new",
                body: JSON.stringify({ name: budgetName })
            });
            const response = await result.json();
            commit("addBudget", response)
        },
        async [SET_CURRENT_BUDGET]({ state, commit, dispatch, rootState }, budgetid) {
            commit("setCurrentBudgetID", budgetid);

            if (budgetid == null)
                return

            await dispatch(FETCH_BUDGET, budgetid)
        },
        async [FETCH_BUDGET]({ state, commit, dispatch, rootState }, budgetid) {
            const result = await dispatch("GET", { path: "/budget/" + budgetid });
            const response = await result.json();
            return commit("setAccounts", response.Accounts);
        },
        async [FETCH_ACCOUNT]({ state, commit, rootState }, accountid) {
            const result = await fetch("/api/v1/account/" + accountid + "/transactions", {
                headers: {
                    'Authorization': 'Bearer ' + rootState.Session.Token
                }
            });
            const response = await result.json();
            commit("setTransactions", response.Transactions);
        },
        async [SET_CURRENT_ACCOUNT]({ state, commit, dispatch, getters }, { budgetid, accountid }) {
            if (budgetid == null)
                return

            await dispatch(FETCH_BUDGET, budgetid);
            commit("setCurrentAccountID", accountid);
            if (accountid == null)
                return

            commit(TITLE, getters.CurrentAccount.Name);
            await dispatch(FETCH_ACCOUNT, accountid)
        }
    },
    getters: {
        Budgets(state) {
            return state.Budgets || [];
        },
        AuthHeaders(state) {
            return {
                'Authorization': 'Bearer ' + state.Session.Token
            }
        },
        CurrentBudget(state) {
            if (state.CurrentBudgetID == null)
                return {};

            return state.Budgets.get(state.CurrentBudgetID);
        },
        Accounts(state) {
            return state.Accounts || [];
        },
        CurrentAccount(state) {
            if (state.CurrentAccountID == null)
                return { name: "Not found" };
            return state.Accounts.get(state.CurrentAccountID);
        },
        OnBudgetAccounts(state) {
            return Array.from(state.Accounts.values()).filter(x => x.OnBudget);
        },
        OffBudgetAccounts(state) {
            return Array.from(state.Accounts.values()).filter(x => !x.OnBudget);
        },
        Transactions(state) {
            return (state.Transactions || []);
        }
    },
    plugins: [createLogger()]
})

store.subscribe((mutation, state) => {
    let persistedState = {
        Session: state.Session,
        Budgets: state.Budgets,
        Accounts: state.Accounts,
        CurrentBudgetID: state.CurrentBudgetID,
        CurrentAccountID: state.CurrentAccountID,
        ShowMenu: state.ShowMenu
    }
    localStorage.setItem("store", JSON.stringify(persistedState));
})